var Moniker = require('moniker')
var fName = Moniker.generator([Moniker.adjective])
var lName = Moniker.generator([Moniker.noun])
const { v4: uuidv4 } = require('uuid')

worldObj = require("../storageObjects/worldObj")
creatureArr = require("../storageObjects/creatureArr")

function percentify(lower,upper,input){
  upper -= lower
  input -= lower
  return input*100/upper
}

class Creature {
  constructor(parentName,surName,location, energy, maxEnergy, speed, brain, brainSize, brainComplexity, owner, activationValue, oscillatorFrequency,maxAge,generation,eatFoodLevel, colour) {
    this._id = uuidv4()
    this.name = fName.choose()
    this.parentName = parentName || this.name
    this.surName = surName || lName.choose()
    this.fullName = this.name+"."+this.parentName+"."+this.surName
    this.maxEnergy = maxEnergy || config.creature.physical.maxEnergy
    this.energy = energy || this.maxEnergy
    this.speed = speed || config.creature.physical.defaultSpeed
    this.brain = brain || []
    this.brainSize = brainSize || config.creature.brain.size
    this.brainComplexity = brainComplexity || config.creature.brain.complexity
    this.activationValue = activationValue || config.creature.physical.activationValue
    this.oscillatorFrequency = oscillatorFrequency || config.creature.brain.oscillatorFrequency
    this.age = 0 
    this.maxAge = maxAge || config.creature.physical.maxAge
    this.generation = generation || 0 
    this.eatFoodLevel = eatFoodLevel || config.creature.physical.eatFoodLevel
    this.kills = 0
    this.births = 0
    this.alive = true
    this.colour = colour || [Math.floor(Math.random()*255),Math.floor(Math.random()*255),Math.floor(Math.random()*255)]
  }
  synapses = {
    input: {
      random:() => {
        let out = Math.random()*100
        return out
      },
      oscillator:() => {
        let x = worldObj.step%this.oscillatorFrequency
         if(x > this.oscillatorFrequency/2){
          x = this.oscillatorFrequency-x
         }
        return x
      },
      energyCheck:() => {
        return percentify(0,this.maxEnergy,this.energy)
      },
      directionCheck:() => {
        switch (this.facingDirection) {
          case 1:
            return 0
          case 2:
            return 25
          case 3:
            return 50
          case 4:
            return 100
        }
      },
    },
    internal: {
      invert:(input) => {
        return input*-1
      },
      halve:(input) => {
        return input*0.5
      },
      double:(input) => {
        return input*2
      },
    },
    action: {
      mate:(input) => {
        if (input > this.activationValue && creatureArr.length < config.maxPop && this.alive && this.energy > (config.creature.physical.birthCost +1)){
          
          // TODO: Figure out why colour Arr sometimes becomes a string.
          if(typeof this.colour == "string"){
            this.die( this.colour)
            return false
          }
          try {
            let obj = this.mutate(JSON.parse(JSON.stringify({
              surName:this.surName,
              maxEnergy:this.maxEnergy,
              speed:this.speed,
              brain:this.brain,
              brainSize:this.brainSize,
              brainComplexity:this.brainComplexity,
              activationValue:this.activationValue,
              oscillatorFrequency: this.oscillatorFrequency,
              maxAge: this.maxAge,
              eatFoodLevel: this.eatFoodLevel,
              colour: this.colour
            })))
            
            if (true){
              this.energy -= config.creature.physical.birthCost
              this.births += 1
              
              let baby = new Creature(this.name,obj.surName,this.behindLoc,config.creature.physical.birthEnergy,obj.maxEnergy,obj.speed,obj.brain,obj.brainSize,obj.brainComplexity,this.owner,obj.activationValue,obj.oscillatorFrequency,obj.maxAge, this.generation+1,obj.eatFoodLevel, obj.colour)
              creatureArr.push(baby)
              return 100
            }
            return 0 
          }
          catch (e){
            this.die(e)
          }
          
        }
        return 0 
      },
    }
  }
  mutate(obj){
    let mod = 0
    if (Math.random() < config.creature.physical.mutateFraction){
      obj.surName = (this.name+"_"+this.surName)
      if(Math.random() > 0.5){
        mod = 1
      } else {
        mod = -1
      }
      let colourSelect = Math.floor(Math.random()*obj.colour.length)
      if (obj.colour[colourSelect] + mod*10 > 255 || obj.colour[colourSelect] + mod*10 < 0){
        obj.colour[colourSelect] += mod*-1*10
        obj.colour[colourSelect] = parseInt(obj.colour[colourSelect])
      } else {
        obj.colour[colourSelect] += mod*10
        obj.colour[colourSelect] = parseInt(obj.colour[colourSelect])
      }
      let pickedNumber = Math.floor(Math.random()*Object.keys(obj).length)
      if (Object.keys(obj)[pickedNumber] != "brain"){
        obj[Object.keys(obj)[pickedNumber]] += mod
      } else {
        let brainPick1 = Math.floor(Math.random()*this.brainSize)
        let brainPick2 = Math.floor(Math.random()*this.brainComplexity)
        let brainPick3 = Math.floor(Math.random()*2)
        if (brainPick3 == 0){
          if (Object.keys(this.synapses).length > (obj.brain[brainPick1][brainPick2][0] + mod) && (obj.brain[brainPick1][brainPick2][0] + mod) > 0 && Object.keys(obj.brain[brainPick1][brainPick2][0]+ mod).length != 0 ){
            obj.brain[brainPick1][brainPick2][0] += mod
          }
        } else {
          if (Object.keys(this.synapses[Object.keys(this.synapses)[obj.brain[brainPick1][brainPick2][0]]]).length > (obj.brain[brainPick1][brainPick2][1] + mod) && (obj.brain[brainPick1][brainPick2][1] + mod) > 0 ){
            obj.brain[brainPick1][brainPick2][1] += mod
          }
        }
      }
    }
    return obj
  }


  randomBrain() {
    for (let braindex = 0; braindex < this.brainSize; braindex++) {
      let brainPath = []
      for (let comlexdex = 0; comlexdex < this.brainComplexity; comlexdex++) {
        let typeNumber=Math.floor(Math.random()*Object.keys(this.synapses).length)
        let actionNumber=Math.floor(Math.random()*Object.keys(this.synapses[Object.keys(this.synapses)[typeNumber]]).length)
        brainPath.push([typeNumber,actionNumber])
      }
      this.brain.push(brainPath)
    }
  }
  act(){
    if (this.brain.length == 0) this.randomBrain()
    if (!this.alive) return 0
    if (this.speed > this.brainSize) this.speed = this.brainSize
    for (let dex = 0; dex < this.speed; dex++){
      let result = 0
      if (typeof this.brain[dex] == "undefined") {
        this.die("their brain size was "+this.brainSize)
        return 0
      } else {
        this.brain[dex].forEach(element => {
          let typeNumber = element[0]
          let actionNumber = element[1]
          let type = Object.keys(this.synapses)[typeNumber]
          let action = null
          action = Object.keys(this.synapses[Object.keys(this.synapses)[typeNumber]])[actionNumber]
          if (this.alive){
            result = this.synapses[type][action](result)
          } else {
            return 0 
          }
          
        })
      }
      }
    this.energy -= 1
    this.age += 1
    if (this.age > this.maxAge){
      this.die("they grew too old")
    }
    if (this.energy <= 0 ){
      this.die("they starved")
    }
    return percentify(0,this.maxEnergy,this.energy)
  }
  die(deathReason){
    if (this.alive){
      this.alive = false
      console.log(this.fullName+" died: " + deathReason)
    }
    return 0 
  }
}

module.exports = Creature