const { v4: uuidv4 } = require('uuid')

userObj = require("../storageObjects/userObj")
worldObj = require("../storageObjects/worldObj")
creatureArr = require("../storageObjects/creatureArr")
mongo = require("../mongoHelp")

class User {
  constructor(name, pass, id, energy, powers){
    this._id = id ||uuidv4()
    this.name = name || "Nobody"
    this.pass = pass
    this.energy = energy || 0
    this.creatures = 0
    this.livingCreatures = 0
    this.deadCreatures = 0
    this.mutants = 0
    this.killedCreatures = 0
    this.readyState = true
    this.powers =  powers || {
      setPlayerPower: [false,0],
      transferEnergy: [true,config.powers.transferEnergyBuyCost],
      saveCreature: [false,config.powers.saveCreatureBuyCost],
      spawnCreature: [false,config.powers.spawnCreatureBuyCost],
      mutateCreature: [false,config.powers.mutateCreatureBuyCost],
      feedCreature: [false,config.powers.feedCreatureBuyCost],
      growFood: [false,config.powers.growFoodBuyCost],
      deleteFood: [false,config.powers.deleteFoodBuyCost],
      bomb: [false,config.powers.bombBuyCost]
    }
    userObj[this.name] = this
    this.initialSaveCheck()
  }
  addEnergy(energy){
    this.energy += energy
    return true
  }
  spendEnergy(energy){
    if ((this.energy - energy) < 0){
      return false
    } else {
      this.energy -= energy
      return true
    }
  }
  addCreature(){
    this.creatures += 1
    this.livingCreatures += 1
  }
  takeCreature(){
    this.deadCreatures += 1
    this.livingCreatures -= 1
  }
  addKill(){
    this.killedCreatures += 1
  }
  addMutant(){
    this.mutants += 1
  }
  bomb(location, radius){
    if(this.powers.bomb[0]){
      let energyCost = (Math.round(radius*3.14)*3000)
      if(this.spendEnergy(energyCost)){
        let count = 0 
        for (let x = 0; x < radius; x++){
          for (let y = 0; y < radius; y++){
            let compx = x
            let compy  = y
            if (x < 0){
              compx *= -1
            }
            if (y < 0){
              compy *= -1
            }
            let listArr = []
            // TODO make the exposions circles and not squares
            // compx+compy <= radius && location[0]+x > 0 && location[1]+y > 0 && location[0]+x < 100 && location[1]+y < 100
            if(true){
              listArr.push([location[0]+x,location[1]+y])
              listArr.push([location[0]+x,location[1]-y])
              listArr.push([location[0]-x,location[1]+y])
              listArr.push([location[0]-x,location[1]-y])
            }
            listArr.forEach(tile => {
              let killedArr = creatureArr.filter(creature => (creature.location[0] == tile[0]) && (creature.location[1] == tile[1]))

              killedArr.forEach(creature => {
                console.log(creature.fullName)
                creature.die("they blew up")
              })
            })
            
            // let dx = x - location[0]
            // let dy = y - location[1]
            // if ((radius*radius) <= (dx * dx + dy * dy)){
            //   console.log(x + y * radius)
            // }
          }

        }
        console.log(count)
      } else return false
    } else return false


  }
  purchasePower(powerName){
    console.log(this.powers[Object.keys(this.powers).filter(power => power == powerName)][1])
    this.spendEnergy(this.powers[Object.keys(this.powers).filter(power => power == powerName)][1])
    this.powers[powerName][0] = true
  }
  clean(){
    this.creatures = 0
    this.livingCreatures = 0
    this.deadCreatures = 0
    this.mutants = 0
    this.killedCreatures = 0
    this.readyState = false
    mongo.update("algoZoo", "users", {_id:this._id}, this)
  }
  initialSaveCheck(){
    mongo.get("algoZoo", "users",{_id: this._id}).then(list => {
      if (list.length == 0){
        mongo.put("algoZoo", "users",this)
      } else {
        mongo.update("algoZoo", "users",{_id: this._id}, this)
      }
    })
  }
}

module.exports = User